"""The main settings module used by caluma service."""

import re

from django.conf import global_settings
from promise import async_instance

from .caluma import *  # noqa
from .caluma import GRAPHENE, default, django_root, env, environ

DEBUG = env.bool("DEBUG", default=default(True, False))
SECRET_KEY = env.str("SECRET_KEY", default=default("uuuuuuuuuu"))
ALLOWED_HOSTS = env.list("ALLOWED_HOSTS", default=default(["*"]))


# Application definition

INSTALLED_APPS = [
    "django.contrib.postgres",
    "localized_fields",
    "psqlextra",
    "django.contrib.contenttypes",
    "django.contrib.staticfiles",
    "graphene_django",
    "simple_history",
    "caluma.caluma_core.apps.DefaultConfig",
    "caluma.caluma_user.apps.DefaultConfig",
    "caluma.caluma_form.apps.DefaultConfig",
    "caluma.caluma_workflow.apps.DefaultConfig",
    "caluma.caluma_data_source.apps.DefaultConfig",
    "caluma.caluma_logging.apps.DefaultConfig",
]

if DEBUG:
    INSTALLED_APPS.append("django_extensions")

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "simple_history.middleware.HistoryRequestMiddleware",
]

ROOT_URLCONF = "caluma.urls"
WSGI_APPLICATION = "caluma.wsgi.application"
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [django_root("caluma", "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]


# Database
# https://docs.djangoproject.com/en/1.11/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "psqlextra.backend",
        "NAME": env.str("DATABASE_NAME", default="caluma"),
        "USER": env.str("DATABASE_USER", default="caluma"),
        "PASSWORD": env.str("DATABASE_PASSWORD", default=default("caluma")),
        "HOST": env.str("DATABASE_HOST", default="localhost"),
        "PORT": env.str("DATABASE_PORT", default=""),
    }
}

# Cache
# https://docs.djangoproject.com/en/1.11/ref/settings/#caches

CACHES = {
    "default": {
        "BACKEND": env.str(
            "CACHE_BACKEND", default="django.core.cache.backends.locmem.LocMemCache"
        ),
        "LOCATION": env.str("CACHE_LOCATION", ""),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/


def parse_languages(languages):
    return [(language, language) for language in languages]


LANGUAGE_CODE = env.str("LANGUAGE_CODE", "en")
LANGUAGES = (
    parse_languages(env.list("LANGUAGES", default=[])) or global_settings.LANGUAGES
)

TIME_ZONE = env.str("TIME_ZONE", "UTC")
USE_I18N = True
USE_L10N = True
USE_TZ = True


def parse_admins(admins):  # pragma: no cover
    """
    Parse env admins to django admins.

    Example of ADMINS environment variable:
    Test Example <test@example.com>,Test2 <test2@example.com>
    """
    result = []
    for admin in admins:
        match = re.search(r"(.+) \<(.+@.+)\>", admin)
        if not match:
            raise environ.ImproperlyConfigured(
                'In ADMINS admin "{0}" is not in correct '
                '"Firstname Lastname <email@example.com>"'.format(admin)
            )
        result.append((match.group(1), match.group(2)))
    return result


ADMINS = parse_admins(env.list("ADMINS", default=[]))


# Cors headers
CORS_ORIGIN_ALLOW_ALL = env.bool("CORS_ORIGIN_ALLOW_ALL", default=False)
CORS_ORIGIN_WHITELIST = env.list("CORS_ORIGIN_WHITELIST", default=[])


# Logging

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {"console": {"class": "logging.StreamHandler"}},
    "loggers": {
        "": {"handlers": ["console"], "level": env.str("LOG_LEVEL", default="INFO")}
    },
}

# static files

STATIC_URL = "/static/"

# disable trampoline to improve performance
# see https://github.com/graphql-python/graphene/issues/268
# this means though that data loader won't work which we currently don't use
# best to remove once graphql-core-next is compatible with graphene.
async_instance.disable_trampoline()


# GraphQL
if DEBUG:
    GRAPHENE["MIDDLEWARE"].append("graphene_django.debug.DjangoDebugMiddleware")


# caluma access log
ENABLE_ACCESS_LOG = env.bool("ENABLE_ACCESS_LOG", default=False)

if ENABLE_ACCESS_LOG:
    MIDDLEWARE.append("caluma.caluma_logging.middleware.AccessLogMiddleware")



from django.core.exceptions import ValidationError
from django.utils import timezone

from caluma.caluma_form.domain_logic import SaveDocumentLogic

from . import api, models, utils, validators
from .events import send_event_with_deprecations


class StartCaseLogic:
    """
    Shared domain logic for starting cases.

    Used in the `saveCase` mutation and in the `start_case` API. The logic
    for case creation is split in three parts (`validate_for_start`,
    `pre_start` and `post_start`) so that in between the appropriate create
    method can be called (`super().create(...)` for the serializer and
    `Case.objects.create(...) for the python API`).
    """

    @staticmethod
    def validate_for_start(data):
        form = data.get("form")
        workflow = data.get("workflow")

        if form:
            if (
                not workflow.allow_all_forms
                and not workflow.allow_forms.filter(pk=form.pk).exists()
            ):
                raise ValidationError(
                    f"Workflow {workflow.pk} does not allow to start case with form {form.pk}"
                )

        return data

    @staticmethod
    def pre_start(validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_create_case",
            sender="pre_create_case",
            case=None,
            validated_data=validated_data,
            user=user,
            context=context,
        )

        parent_work_item = validated_data.get("parent_work_item")
        validated_data["status"] = models.Case.STATUS_RUNNING

        form = validated_data.pop("form", None)
        if form:
            validated_data["document"] = SaveDocumentLogic.create(
                {"form": form}, user=user
            )

        if parent_work_item:
            case = parent_work_item.case
            while hasattr(case, "parent_work_item"):
                case = case.parent_work_item.case
            validated_data["family"] = case

        return validated_data

    @staticmethod
    def post_start(case, user, parent_work_item, context=None):
        # Django doesn't save reverse one-to-one relationships automatically:
        # https://code.djangoproject.com/ticket/18638
        if parent_work_item:
            parent_work_item.child_case = case
            parent_work_item.save()

        workflow = case.workflow
        tasks = workflow.start_tasks.all()

        work_items = utils.bulk_create_work_items(tasks, case, user, None, context)

        send_event_with_deprecations(
            "post_create_case",
            sender="case_post_create",
            case=case,
            user=user,
            context=context,
        )

        for work_item in work_items:  # pragma: no cover
            send_event_with_deprecations(
                "post_create_work_item",
                sender="case_post_create",
                work_item=work_item,
                user=user,
                context=context,
            )

        return case


class CompleteWorkItemLogic:
    @staticmethod
    def _can_continue(work_item, task):
        # If a "multiple instance" task has running siblings, the task is not completed
        if task.is_multiple_instance:
            return not work_item.case.work_items.filter(
                task=task, status=models.WorkItem.STATUS_READY
            ).exists()
        return work_item.case.work_items.filter(
            task=task,
            status__in=(
                models.WorkItem.STATUS_COMPLETED,
                models.WorkItem.STATUS_SKIPPED,
            ),
        ).exists()

    @staticmethod
    def validate_for_complete(work_item, user):
        validators.WorkItemValidator().validate(
            status=work_item.status,
            child_case=work_item.child_case,
            task=work_item.task,
            case=work_item.case,
            document=work_item.document,
            user=user,
        )

    @staticmethod
    def pre_complete(
        work_item,
        validated_data,
        user,
        context=None,
        event="pre_complete_work_item",
        sender="pre_complete_work_item",
    ):
        send_event_with_deprecations(
            event, sender=sender, work_item=work_item, user=user, context=context
        )

        validated_data["status"] = models.WorkItem.STATUS_COMPLETED
        validated_data["closed_at"] = timezone.now()
        validated_data["closed_by_user"] = user.username
        validated_data["closed_by_group"] = user.group

        return validated_data

    @staticmethod
    def post_complete(
        work_item,
        user,
        context=None,
        event="post_complete_work_item",
        sender="post_complete_work_item",
    ):
        case = work_item.case

        send_event_with_deprecations(
            event, sender=sender, work_item=work_item, user=user, context=context
        )

        if not CompleteWorkItemLogic._can_continue(work_item, work_item.task):
            return work_item

        flow = models.Flow.objects.filter(
            task_flows__task=work_item.task_id,
            task_flows__workflow=work_item.case.workflow.pk,
        ).first()

        next_tasks = models.Task.objects.filter(
            pk__in=utils.get_jexl_tasks(
                flow.next if flow else None, work_item.case, user, work_item, context
            )
        )

        if next_tasks.exists():
            sibling_tasks = models.Task.objects.filter(task_flows__flow=flow)

            all_siblings_complete = all(
                CompleteWorkItemLogic._can_continue(work_item, task)
                for task in sibling_tasks
            )

            if all_siblings_complete:
                created_work_items = utils.bulk_create_work_items(
                    next_tasks, case, user, work_item, context
                )

                for created_work_item in created_work_items:  # pragma: no cover
                    send_event_with_deprecations(
                        "post_create_work_item",
                        sender=sender,
                        work_item=created_work_item,
                        user=user,
                        context=context,
                    )

        if (
            not next_tasks.exists()
            and not work_item.case.work_items.filter(
                status=models.WorkItem.STATUS_READY
            ).exists()
        ):
            # no more tasks, mark case as complete
            case.status = models.Case.STATUS_COMPLETED
            case.closed_at = timezone.now()
            case.closed_by_user = user.username
            case.closed_by_group = user.group
            case.save()

            send_event_with_deprecations(
                "post_complete_case",
                sender=sender,
                case=case,
                user=user,
                context=context,
            )

        return work_item


class SkipWorkItemLogic:
    @staticmethod
    def validate_for_skip(work_item):
        if work_item.case.status != models.Case.STATUS_RUNNING:
            raise ValidationError("Only work items of running cases can be skipped.")

        if work_item.status != models.WorkItem.STATUS_READY:
            raise ValidationError("Only ready work items can be skipped.")

    @staticmethod
    def pre_skip(work_item, validated_data, user, context=None):
        validated_data = CompleteWorkItemLogic.pre_complete(
            work_item,
            validated_data,
            user,
            context,
            event="pre_skip_work_item",
            sender="pre_skip_work_item",
        )

        validated_data["status"] = models.WorkItem.STATUS_SKIPPED

        if (
            work_item.child_case
            and work_item.child_case.status == models.Case.STATUS_RUNNING
        ):
            api.cancel_case(work_item.child_case, user, context)

        return validated_data

    @staticmethod
    def post_skip(work_item, user, context=None):
        work_item = CompleteWorkItemLogic.post_complete(
            work_item,
            user,
            context,
            event="post_skip_work_item",
            sender="post_skip_work_item",
        )

        return work_item


class CancelCaseLogic:
    """
    Shared domain logic for canceling cases.

    Used in the `cancelCase` mutation and in the `cancel_case` API. The logic
    for case cancelation is split in three parts (`validate_for_cancel`,
    `pre_cancel` and `post_cancel`) so that in between the appropriate update
    method can be called (`super().update(...)` for the serializer and
    `Case.objects.update(...) for the python API`).
    """

    @staticmethod
    def validate_for_cancel(case):
        if case.status not in [
            models.Case.STATUS_RUNNING,
            models.Case.STATUS_SUSPENDED,
        ]:
            raise ValidationError("Only running or suspended cases can be canceled.")

    @staticmethod
    def pre_cancel(case, validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_cancel_case",
            sender="pre_cancel_case",
            case=case,
            user=user,
            context=context,
        )

        validated_data["status"] = models.Case.STATUS_CANCELED
        validated_data["closed_at"] = timezone.now()
        validated_data["closed_by_user"] = user.username
        validated_data["closed_by_group"] = user.group

        for work_item in case.work_items.filter(
            status__in=[models.WorkItem.STATUS_READY, models.WorkItem.STATUS_SUSPENDED]
        ):
            api.cancel_work_item(work_item, user, context)

        return validated_data

    @staticmethod
    def post_cancel(case, user, context=None):
        send_event_with_deprecations(
            "post_cancel_case",
            sender="post_cancel_case",
            case=case,
            user=user,
            context=context,
        )

        return case


class SuspendCaseLogic:
    """
    Shared domain logic for suspending cases.

    Used in the `suspendCase` mutation and in the `suspend_case` API. The logic
    for case suspension is split in three parts (`validate_for_suspend`,
    `pre_suspend` and `post_suspend`) so that in between the appropriate update
    method can be called (`super().update(...)` for the serializer and
    `Case.objects.update(...) for the python API`).
    """

    @staticmethod
    def validate_for_suspend(case):
        if case.status != models.Case.STATUS_RUNNING:
            raise ValidationError("Only running cases can be suspended.")

    @staticmethod
    def pre_suspend(case, validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_suspend_case",
            sender="pre_suspend_case",
            case=case,
            user=user,
            context=context,
        )

        validated_data["status"] = models.Case.STATUS_SUSPENDED

        for work_item in case.work_items.filter(status=models.WorkItem.STATUS_READY):
            api.suspend_work_item(work_item, user, context)

        return validated_data

    @staticmethod
    def post_suspend(case, user, context=None):
        send_event_with_deprecations(
            "post_suspend_case",
            sender="post_suspend_case",
            case=case,
            user=user,
            context=context,
        )

        return case


class ResumeCaseLogic:
    """
    Shared domain logic for resuming cases.

    Used in the `resumeCase` mutation and in the `resume_case` API. The logic
    for resuming a case is split in three parts (`validate_for_resume`,
    `pre_resume` and `post_resume`) so that in between the appropriate update
    method can be called (`super().update(...)` for the serializer and
    `Case.objects.update(...) for the python API`).
    """

    @staticmethod
    def validate_for_resume(case):
        if case.status != models.Case.STATUS_SUSPENDED:
            raise ValidationError("Only suspended cases can be resumed.")

    @staticmethod
    def pre_resume(case, validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_resume_case",
            sender="pre_resume_case",
            case=case,
            user=user,
            context=context,
        )

        validated_data["status"] = models.Case.STATUS_RUNNING

        for work_item in case.work_items.filter(
            status=models.WorkItem.STATUS_SUSPENDED
        ):
            api.resume_work_item(work_item, user, context)

        return validated_data

    @staticmethod
    def post_resume(case, user, context=None):
        send_event_with_deprecations(
            "post_resume_case",
            sender="post_resume_case",
            case=case,
            user=user,
            context=context,
        )

        return case


class CancelWorkItemLogic:
    @staticmethod
    def validate_for_cancel(work_item):
        if work_item.status not in [
            models.WorkItem.STATUS_READY,
            models.WorkItem.STATUS_SUSPENDED,
        ]:
            raise ValidationError("Only ready or suspended work items can be canceled.")

    @staticmethod
    def pre_cancel(work_item, validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_cancel_work_item",
            sender="pre_cancel_work_item",
            work_item=work_item,
            user=user,
            context=context,
        )

        validated_data["status"] = models.WorkItem.STATUS_CANCELED
        validated_data["closed_at"] = timezone.now()
        validated_data["closed_by_user"] = user.username
        validated_data["closed_by_group"] = user.group

        if work_item.child_case and work_item.child_case.status in [
            models.Case.STATUS_RUNNING,
            models.Case.STATUS_SUSPENDED,
        ]:
            api.cancel_case(work_item.child_case, user, context)

        return validated_data

    @staticmethod
    def post_cancel(work_item, user, context=None):
        send_event_with_deprecations(
            "post_cancel_work_item",
            sender="post_cancel_work_item",
            work_item=work_item,
            user=user,
            context=context,
        )

        return work_item


class SuspendWorkItemLogic:
    @staticmethod
    def validate_for_suspend(work_item):
        if work_item.status != models.WorkItem.STATUS_READY:
            raise ValidationError("Only ready work items can be suspended.")

    @staticmethod
    def pre_suspend(work_item, validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_suspend_work_item",
            sender="pre_suspend_work_item",
            work_item=work_item,
            user=user,
            context=context,
        )

        validated_data["status"] = models.WorkItem.STATUS_SUSPENDED

        if (
            work_item.child_case
            and work_item.child_case.status == models.Case.STATUS_RUNNING
        ):
            api.suspend_case(work_item.child_case, user, context)

        return validated_data

    @staticmethod
    def post_suspend(work_item, user, context=None):
        send_event_with_deprecations(
            "post_suspend_work_item",
            sender="post_suspend_work_item",
            work_item=work_item,
            user=user,
            context=context,
        )

        return work_item


class ResumeWorkItemLogic:
    @staticmethod
    def validate_for_resume(work_item):
        if work_item.status != models.WorkItem.STATUS_SUSPENDED:
            raise ValidationError("Only suspended work items can be resumed.")

    @staticmethod
    def pre_resume(work_item, validated_data, user, context=None):
        send_event_with_deprecations(
            "pre_resume_work_item",
            sender="pre_resume_work_item",
            work_item=work_item,
            user=user,
            context=context,
        )

        validated_data["status"] = models.WorkItem.STATUS_READY

        if (
            work_item.child_case
            and work_item.child_case.status == models.Case.STATUS_SUSPENDED
        ):
            api.resume_case(work_item.child_case, user, context)

        return validated_data

    @staticmethod
    def post_resume(work_item, user, context=None):
        send_event_with_deprecations(
            "post_resume_work_item",
            sender="post_resume_work_item",
            work_item=work_item,
            user=user,
            context=context,
        )

        return work_item



import pytest


@pytest.mark.parametrize(
    "factory,query,reverse",
    [
        ("case_factory", "allCases", True),
        ("case_factory", "allCases", False),
        ("document_factory", "allDocuments", True),
        ("document_factory", "allDocuments", False),
    ],
)
def test_order_by_meta(db, schema_executor, request, factory, query, reverse):

    factory = request.getfixturevalue(factory)
    factory(meta={"test-key": 3})
    factory(meta={"test-key": 100})
    factory(meta={"test-key": 20})

    order = "META_TEST_KEY_DESC" if reverse else "META_TEST_KEY_ASC"

    resp = schema_executor(
        f"""
            query {{
                {query} (orderBy: {order}) {{
                    edges {{
                        node {{
                            meta
                        }}
                    }}
                }}
            }}
        """
    )

    assert not resp.errors

    edges = resp.data[query]["edges"]

    assert edges == sorted(
        edges, key=lambda e: e["node"]["meta"]["test-key"], reverse=reverse
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-29 09:10
from __future__ import unicode_literals

import localized_fields.fields.text_field
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0013_auto_20190418_0733")]

    operations = [
        migrations.AddField(
            model_name="question",
            name="static_content",
            field=localized_fields.fields.text_field.LocalizedTextField(
                blank=True, null=True, required=[]
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                    ("file", "file"),
                    ("dynamic_choice", "dynamic_choice"),
                    ("dynamic_multiple_choice", "dynamic_multiple_choice"),
                    ("static", "static"),
                ],
                max_length=23,
            ),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-31 08:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0025_dynamicoption_historicaldynamicoption")]

    operations = [
        migrations.RemoveField(model_name="dynamicoption", name="value"),
        migrations.RemoveField(model_name="historicaldynamicoption", name="value"),
        migrations.AddField(
            model_name="dynamicoption",
            name="slug",
            field=models.CharField(default="unset", max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="historicaldynamicoption",
            name="slug",
            field=models.CharField(default="unset", max_length=255),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.10 on 2020-03-19 09:15

import localized_fields.fields.field
from django.db import migrations


def set_name_and_description(apps, schema_editor):
    WorkItem = apps.get_model("caluma_workflow", "WorkItem")
    HistoricalWorkItem = apps.get_model("caluma_workflow", "HistoricalWorkItem")
    db_alias = schema_editor.connection.alias

    for work_item in WorkItem.objects.using(db_alias).iterator():
        work_item.name = work_item.task.name
        work_item.description = work_item.task.description
        work_item.save()
        for historical_work_item in HistoricalWorkItem.objects.using(db_alias).filter(
            id=work_item.pk
        ):
            historical_work_item.name = work_item.name
            historical_work_item.description = work_item.description
            historical_work_item.save()


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0021_work_item_controlling_groups")]

    operations = [
        migrations.AddField(
            model_name="historicalworkitem",
            name="description",
            field=localized_fields.fields.field.LocalizedField(
                blank=True,
                help_text="Will be set from Task, if not provided.",
                null=True,
                required=[],
            ),
        ),
        migrations.AddField(
            model_name="historicalworkitem",
            name="name",
            field=localized_fields.fields.field.LocalizedField(
                default={"de": "temp_placeholder"},
                help_text="Will be set from Task, if not provided.",
                required=[],
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="workitem",
            name="description",
            field=localized_fields.fields.field.LocalizedField(
                blank=True,
                help_text="Will be set from Task, if not provided.",
                null=True,
                required=[],
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="name",
            field=localized_fields.fields.field.LocalizedField(
                default={"de": "temp_placeholder"},
                help_text="Will be set from Task, if not provided.",
                required=[],
            ),
            preserve_default=False,
        ),
        migrations.RunPython(set_name_and_description, migrations.RunPython.noop),
    ]



import os

import pytest
from graphql_relay import to_global_id

from ..schema import schema

DIR_NAME = os.path.dirname(__file__)


@pytest.mark.parametrize("node_type", ["form", "workflow"])
def test_schema_node(db, snapshot, request, node_type):
    """
    Add your model to parametrize for automatic global node testing.

    Requirement is that node and model have the same name
    """
    node_instance = request.getfixturevalue(node_type)
    global_id = to_global_id(node_instance.__class__.__name__, node_instance.pk)

    node_query = """
    query %(name)s($id: ID!) {
      node(id: $id) {
        ... on %(name)s {
          id
        }
      }
    }
    """ % {
        "name": node_instance.__class__.__name__
    }

    result = schema.execute(node_query, variable_values={"id": global_id})
    assert not result.errors
    assert result.data["node"]["id"] == global_id


def test_schema_introspect(snapshot):
    snapshot.assert_match(str(schema))



from typing import Any, Optional

from caluma.caluma_form import domain_logic, models
from caluma.caluma_user.models import BaseUser


def save_answer(
    question: models.Question,
    document: Optional[models.Document] = None,
    user: Optional[BaseUser] = None,
    value: Optional[Any] = None,
    context: Optional[dict] = None,
    **kwargs
) -> models.Answer:
    """
    Save an answer for given question, document.

    Similar to saveDocumentStringAnswer and the likes, it performes upsert.
    :param value: Must match the question type
    """

    data = {"question": question, "document": document, "value": value}
    data.update(kwargs)

    answer = models.Answer.objects.filter(question=question, document=document).first()
    answer = domain_logic.SaveAnswerLogic.get_new_answer(data, user, answer)

    return answer


def save_default_answer(
    question: models.Question,
    user: Optional[BaseUser] = None,
    value: Optional[Any] = None,
    **kwargs
) -> models.Answer:
    """
    Save default_answer for given question.

    Similar to saveDefaultStringAnswer and the likes, it performes upsert.
    :param value: Must match the question type
    """

    data = {"question": question, "value": value}
    data.update(kwargs)

    answer = question.default_answer
    answer = domain_logic.SaveDefaultAnswerLogic.get_new_answer(data, user, answer)

    return answer


def save_document(
    form: models.Form,
    meta: Optional[dict] = None,
    document: Optional[models.Document] = None,
    user: Optional[BaseUser] = None,
) -> models.Document:
    """Save a document for a given form."""

    if meta is None:
        meta = {}

    if not document:
        return domain_logic.SaveDocumentLogic.create(
            {"form": form, "meta": meta}, user=user
        )

    domain_logic.SaveDocumentLogic.update(
        document, {"form": form, "meta": meta}, user=user
    )
    return document


def copy_form(
    source: models.Form,
    slug: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    is_published: Optional[bool] = None,
    user: Optional[BaseUser] = None,
) -> models.Form:
    """Copy a form."""

    return domain_logic.CopyFormLogic.copy(
        {
            "source": source,
            "slug": slug,
            "name": name,
            "description": description,
            "is_published": False if is_published is None else is_published,
        },
        user=user,
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-02-13 10:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0009_auto_20190213_0939")]

    operations = [
        migrations.AddField(
            model_name="task",
            name="lead_time",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="Time in seconds task may take to be processed.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="deadline",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



import sys

from django.dispatch import Signal

from caluma.caluma_core.events import send_event

module = sys.modules[__name__]

MODEL_ACTIONS = {
    "work_item": ["create", "complete", "cancel", "skip", "suspend", "resume"],
    "case": ["create", "complete", "cancel", "suspend", "resume"],
}

# create "pre|post"_"create|complete|..."_"work_item|case" events in all given combinations
for obj, actions in MODEL_ACTIONS.items():
    for action in actions:
        for prefix in ["pre", "post"]:
            providing_args = [obj, "user", "context"]

            if prefix == "pre" and action == "create":
                providing_args += ["validated_data"]

            setattr(
                module,
                f"{prefix}_{action}_{obj}",
                Signal(providing_args=providing_args),
            )

# TODO: remove in the next major release since those are events are deprecated...
# ... in favor of "post_" events
created_work_item = Signal(providing_args=["work_item", "user", "context"])
completed_work_item = Signal(providing_args=["work_item", "user", "context"])
skipped_work_item = Signal(providing_args=["work_item", "user", "context"])
suspended_work_item = Signal(providing_args=["work_item", "user", "context"])
resumed_work_item = Signal(providing_args=["work_item", "user", "context"])
completed_case = Signal(providing_args=["case", "user", "context"])
created_case = Signal(providing_args=["case", "user", "context"])
suspended_case = Signal(providing_args=["case", "user", "context"])
resumed_case = Signal(providing_args=["case", "user", "context"])

# ... also in favor of the american notation
canceled_work_item = Signal(providing_args=["work_item", "user", "context"])
cancelled_work_item = Signal(providing_args=["work_item", "user", "context"])
canceled_case = Signal(providing_args=["case", "user", "context"])
cancelled_case = Signal(providing_args=["case", "user", "context"])

DEPRECATIONS = {
    "post_create_work_item": ["created_work_item"],
    "post_complete_work_item": ["completed_work_item"],
    "post_cancel_work_item": ["canceled_work_item", "cancelled_work_item"],
    "post_skip_work_item": ["skipped_work_item"],
    "post_suspend_work_item": ["suspended_work_item"],
    "post_resume_work_item": ["resumed_work_item"],
    "post_complete_case": ["completed_case"],
    "post_cancel_case": ["canceled_case", "cancelled_case"],
    "post_create_case": ["created_case"],
    "post_suspend_case": ["suspended_case"],
    "post_resume_case": ["resumed_case"],
}

for new, deprecations in DEPRECATIONS.items():
    for old in deprecations:
        getattr(
            module, old
        )._deprecation_reason = f"`caluma_workflow.events.{old}` is deprecated in favor of `caluma_workflow.events.{new}` and will be removed in the next major version."


# TODO remove this in next major release and replace call sites with regular send_event
def send_event_with_deprecations(event_name, **kwargs):
    new_event = getattr(module, event_name)
    send_event(new_event, **kwargs)

    if event_name not in DEPRECATIONS:
        return

    for deprecation in DEPRECATIONS[event_name]:
        old_event = getattr(module, deprecation)
        send_event(old_event, **kwargs)



from factory import Faker, LazyAttribute, django


class DjangoModelFactory(django.DjangoModelFactory):
    created_by_user = Faker("uuid4")
    created_by_group = Faker("uuid4")
    modified_by_user = LazyAttribute(lambda inst: inst.created_by_user)
    modified_by_group = LazyAttribute(lambda inst: inst.created_by_group)



import pytest

from ...caluma_core.relay import extract_global_id
from ...caluma_form.filters import AnswerHierarchyMode, AnswerLookupMode
from .. import models

TEST_VALUES = {
    "multiple_choice": {"matching": ["a", "b"], "nomatch": ["x"]},
    "integer": {"matching": 10, "nomatch": 99},
    "text": {"matching": "foo", "nomatch": "bar"},
    "textarea": {"matching": "foo", "nomatch": "bar"},
    "float": {"matching": 11.5, "nomatch": 99.5},
    "datetime": {
        "matching": "2018-05-09T14:54:51.728786",
        "nomatch": "2019-05-09T14:54:51.728786",
    },
    "date": {"matching": "2018-05-09", "nomatch": "2019-05-09"},
    "choice": {"matching": "a", "nomatch": "x"},
    "calculated_float": {"matching": 11.5, "nomatch": 99.5},
}


@pytest.mark.parametrize(
    "search_lookup", [None, *AnswerLookupMode._meta.enum.__members__]
)
@pytest.mark.parametrize("form_value", ["matching", "nomatch"])
@pytest.mark.parametrize(
    "search_slug,search_value",
    [(qtype, vals["matching"]) for qtype, vals in TEST_VALUES.items()],
)
def test_query_all_questions(
    schema_executor,
    db,
    snapshot,
    question,
    form_factory,
    form_question_factory,
    question_factory,
    document_factory,
    answer_factory,
    form_value,
    search_slug,
    search_value,
    search_lookup,
):
    # build up our form first
    top_form = form_factory()
    top_question = form_question_factory.create(
        form=top_form,
        question__slug="top_question",
        question__type=models.Question.TYPE_TEXT,
    ).question
    subform_question = form_question_factory.create(
        form=top_form,
        question__slug="subform-question",
        question__type=models.Question.TYPE_FORM,
        question__sub_form__slug="subform",
    ).question

    subform_document = document_factory(form=subform_question.sub_form)

    # "randomly" set the hierarchy lookup mode. Not parametrized, as it would
    # explode the test run time three-fold
    possible_hierarchy_lookups = [None, *AnswerHierarchyMode._meta.enum.__members__]
    hierarchy_lookup = possible_hierarchy_lookups[
        len(search_slug) % len(possible_hierarchy_lookups)
    ]

    sub_questions = {
        qtype.lower(): form_question_factory(
            question__slug=qtype.lower(),
            question__type=qtype,
            form=subform_question.sub_form,
        ).question
        for qtype in models.Question.TYPE_CHOICES
        if qtype
        not in (
            # some question types are not searchable
            models.Question.TYPE_FILE,
            models.Question.TYPE_TABLE,
            models.Question.TYPE_FORM,
            models.Question.TYPE_DYNAMIC_CHOICE,
            models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            models.Question.TYPE_STATIC,
        )
    }

    # Now, make a document for it

    result = schema_executor(
        """
            mutation foo ($form_id: ID!){
                saveDocument(input: {form: $form_id}) {
                    document {
                      id
                    }
                }
            }
        """,
        variable_values={"form_id": top_form.pk},
    )

    document_id = extract_global_id(result.data["saveDocument"]["document"]["id"])
    document = models.Document.objects.get(pk=document_id)
    # just some other answer
    answer_factory(question=top_question, document=document, value="foo")

    for qtype, question in sub_questions.items():
        if qtype == models.Question.TYPE_DATE:
            answer_factory(
                question=question,
                document=subform_document,
                date=TEST_VALUES[qtype][form_value],
            )
        elif qtype == models.Question.TYPE_CALCULATED_FLOAT:
            answer = models.Answer.objects.get(
                question__type=models.Question.TYPE_CALCULATED_FLOAT
            )
            answer.value = TEST_VALUES[qtype][form_value]
            answer.save()
        else:
            answer_factory(
                question=question,
                document=subform_document,
                value=TEST_VALUES[qtype][form_value],
            )

    query = """
        query asdf ($hasAnswer: [HasAnswerFilterType]!) {
          allDocuments(hasAnswer: $hasAnswer) {
            edges {
              node {
                form {
                  slug
                }
              }
            }
          }
        }
    """

    variables = {"hasAnswer": [{"question": search_slug, "value": search_value}]}
    if search_lookup is not None:
        variables["hasAnswer"][0]["lookup"] = search_lookup
    if hierarchy_lookup is not None:
        variables["hasAnswer"][0]["hierarchy"] = hierarchy_lookup

    result = schema_executor(query, variable_values=variables)
    # note: result.errors may contain stuff that pytest-snapshottest cannot
    # serialize properly, thus we stringify it
    snapshot.assert_match(
        {
            "request": {"query": query, "variables": variables},
            "response": {"errors": str(result.errors), "data": result.data},
        }
    )


@pytest.mark.parametrize(
    "question__type,value,search,expect_find",
    [
        (models.Question.TYPE_MULTIPLE_CHOICE, ["a", "b"], ["a"], True),
        (models.Question.TYPE_MULTIPLE_CHOICE, ["a", "b"], ["c"], False),
        (models.Question.TYPE_CHOICE, "foo", ["foo", "bar"], True),
        (models.Question.TYPE_CHOICE, "foo", ["bar", "baz"], False),
        (models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, ["a", "b"], ["a"], True),
        (models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, ["a", "b"], ["c"], False),
        (models.Question.TYPE_DYNAMIC_CHOICE, "foo", ["foo", "bar"], True),
        (models.Question.TYPE_DYNAMIC_CHOICE, "foo", ["bar", "baz"], False),
    ],
)
def test_has_answer_intersect(
    schema_executor,
    db,
    question,
    document_factory,
    answer_factory,
    value,
    search,
    expect_find,
):
    doc = document_factory()
    answer_factory(document=doc, question=question, value=value)

    query = """
        query asdf ($hasAnswer: [HasAnswerFilterType]!) {
          allDocuments(hasAnswer: $hasAnswer) {
            edges {
              node {
                id
              }
            }
          }
        }
    """
    variables = {
        "hasAnswer": [
            {"question": question.slug, "value": search, "lookup": "INTERSECTS"}
        ]
    }

    result = schema_executor(query, variable_values=variables)
    assert not result.errors

    expect_count = 1 if expect_find else 0

    assert len(result.data["allDocuments"]["edges"]) == expect_count


@pytest.mark.parametrize("question__is_hidden,expected", [("false", 1), ("true", 0)])
def test_visible_in_context(
    schema_executor, db, document, form, question, form_question, answer, expected
):
    query = """
        query asdf ($visible: Boolean!) {
          allDocuments {
            edges {
              node {
                answers(visibleInContext: $visible) {
                  edges {
                    node {
                      id
                    }
                  }
                }
              }
            }
          }
        }
    """
    variables = {"visible": True}

    result = schema_executor(query, variable_values=variables)
    assert not result.errors
    assert (
        len(result.data["allDocuments"]["edges"][0]["node"]["answers"]["edges"])
        == expected
    )



from django.core.management.commands import migrate

from .migrate_to_prefixed_apps import Command as PrefixCommand


class Command(migrate.Command):
    """migrate command wrapper to add custom logic."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _set_app_prefix(self):
        prefix_migrator = PrefixCommand()
        prefix_migrator.collect()

        if prefix_migrator.queries:  # pragma: no cover
            # this is already tested in test_migrate_to_prefixed_apps
            self.stderr.write(
                "Adding table prefix and updating django internal tables."
            )
            prefix_migrator.apply()

    def handle(self, *args, **options):
        self._set_app_prefix()

        super().handle(*args, **options)



import pytest
from graphql_relay import to_global_id

from caluma.caluma_core.tests import extract_serializer_input_fields

from .. import models, serializers


def test_query_all_workflows(
    db,
    snapshot,
    workflow,
    task_flow,
    workflow_start_tasks,
    workflow_allow_forms,
    flow,
    schema_executor,
):
    query = """
        query AllWorkflows($name: String!) {
          allWorkflows(name: $name) {
            totalCount
            edges {
              node {
                slug
                name
                description
                meta
                tasks {
                  slug
                }
                flows {
                  totalCount
                  edges {
                    node {
                      tasks {
                        slug
                      }
                      next
                    }
                  }
                }
                startTasks {
                    slug
                }
                allowAllForms
                allowForms {
                  edges {
                    node {
                      slug
                    }
                  }
                }
              }
            }
          }
        }
    """

    result = schema_executor(query, variable_values={"name": workflow.name})

    assert not result.errors
    snapshot.assert_match(result.data)


def test_save_workflow(
    db, snapshot, workflow, workflow_start_tasks, workflow_allow_forms, schema_executor
):
    query = """
        mutation SaveWorkflow($input: SaveWorkflowInput!) {
          saveWorkflow(input: $input) {
            workflow {
              allowAllForms
              allowForms {
                totalCount
                edges {
                  node {
                    slug
                  }
                }
              }
              slug
              name
              meta
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveWorkflowSerializer, workflow
        )
    }
    workflow.delete()  # test creation
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "task__slug,next,success",
    [
        ("task-slug", '"task-slug"|task', True),
        ("task-slug", '"not-a-task-slug"|task', False),
        ("task-slug", '["not-a-task-slug"]|tasks', False),
        ("task-slug", '"not-a-task-slug"|invalid', False),
        ("task-slug", '""', False),
    ],
)
def test_add_workflow_flow(
    db,
    workflow,
    workflow_start_tasks_factory,
    task,
    sorted_snapshot,
    success,
    next,
    admin_schema_executor,
):
    workflow_start_tasks_factory.create_batch(2, workflow=workflow)
    query = """
        mutation AddWorkflowFlow($input: AddWorkflowFlowInput!) {
          addWorkflowFlow(input: $input) {
            workflow {
              tasks {
                slug
              }
              flows {
                totalCount
                edges {
                  node {
                    createdByUser
                    createdByGroup
                    next
                    tasks {
                      slug
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    result = admin_schema_executor(
        query,
        variable_values={
            "input": {
                "workflow": to_global_id(type(workflow).__name__, workflow.pk),
                "tasks": [to_global_id(type(task).__name__, task.pk)],
                "next": next,
            }
        },
    )
    assert not bool(result.errors) == success
    if success:
        assert result.data == sorted_snapshot("tasks", lambda el: el["slug"])


def test_remove_flow(db, workflow, task_flow, flow, schema_executor):
    query = """
        mutation RemoveFlow($input: RemoveFlowInput!) {
          removeFlow(input: $input) {
            flow {
              next
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(query, variable_values={"input": {"flow": str(flow.pk)}})
    assert not result.errors
    assert workflow.task_flows.count() == 0


def test_add_workflow_flow_reuse_task(
    db, workflow_start_tasks_factory, task_factory, snapshot, admin_schema_executor
):
    """Test that the same tasks can be reused in multiple flows."""

    task_a, task_b = task_factory.create_batch(2, type=models.Task.TYPE_SIMPLE)
    workflow_a = workflow_start_tasks_factory(
        task=task_a, workflow__slug="workflow-a"
    ).workflow
    workflow_b = workflow_start_tasks_factory(
        task=task_b, workflow__slug="workflow-b"
    ).workflow

    query = """
    mutation AddWorkflowFlow($input: AddWorkflowFlowInput!) {
      addWorkflowFlow(input: $input) {
        clientMutationId
      }
    }
    """

    result = admin_schema_executor(
        query,
        variable_values={
            "input": {
                "workflow": workflow_a.slug,
                "tasks": [task_a.slug],
                "next": f"'{task_b.slug}'|task",
            }
        },
    )
    assert not result.errors

    result = admin_schema_executor(
        query,
        variable_values={
            "input": {
                "workflow": workflow_b.slug,
                "tasks": [task_a.slug],
                "next": f"'{task_b.slug}'|task",
            }
        },
    )
    assert not result.errors

    task_flows = models.TaskFlow.objects.all()
    assert task_flows.count() == 2
    assert (
        task_flows.get(workflow=workflow_a).flow
        != task_flows.get(workflow=workflow_b).flow
    )
    assert (
        task_flows.get(workflow=workflow_a).flow.next
        == task_flows.get(workflow=workflow_b).flow.next
    )



import functools

import pyjexl
import pytest

from ..jexl import JEXL, Cache, CalumaAnalyzer, ExtractTransformSubjectAnalyzer


@pytest.mark.parametrize(
    "expression, expected_transforms",
    [
        ('63 > 62 ? "test"|transform1 : "test2"|transform2', {"test"}),
        ('63 > 62 ? "test"|transform1 : "test2"|transform1', {"test", "test2"}),
        ('"test2"|transform2', set()),
        ('"test2"|transform1', {"test2"}),
    ],
)
def test_extract_transforms(expression, expected_transforms):
    jexl = pyjexl.JEXL()
    jexl.add_transform("transform1", lambda x: x)
    jexl.add_transform("transform2", lambda x: x)

    assert (
        set(
            jexl.analyze(
                expression,
                functools.partial(
                    ExtractTransformSubjectAnalyzer, transforms=["transform1"]
                ),
            )
        )
        == set(expected_transforms)
    )


@pytest.mark.parametrize(
    "expression, expected_log, expected_output",
    [
        ("'foo'", [], "foo"),
        (
            "'foo'|debug",
            ["JEXL debug: in expression `'foo'|debug`, value = `foo`"],
            "foo",
        ),
        (
            "'foo'|debug|debug",
            [
                "JEXL debug: in expression `'foo'|debug|debug`, value = `foo`",
                "JEXL debug: in expression `'foo'|debug|debug`, value = `foo`",
            ],
            "foo",
        ),
        (
            "'\"bar\"|debug'|sub_expr|debug",
            [
                'JEXL debug: in expression `"bar"|debug`, value = `bar`',
                "JEXL debug: in expression `'\"bar\"|debug'|sub_expr|debug`, value = `bar`",
            ],
            "bar",
        ),
        (
            "'hello world'|debug('The world debugger')",
            ["JEXL debug (The world debugger): value = `hello world`"],
            "hello world",
        ),
    ],
)
def test_debug_transform(expression, expected_log, expected_output, caplog):
    jexl = JEXL()

    def sub_expr(val):
        return jexl.evaluate(val)

    jexl.add_transform("sub_expr", sub_expr)

    assert jexl.evaluate(expression) == expected_output
    assert caplog.messages == expected_log


def test_jexl_cache():
    cache = Cache(20, 10)

    # fill the cache "to the brim"
    for x in range(19):
        cache.get_or_set(x, lambda: x)
    assert len(cache._mru) == 19
    assert len(cache._cache) == 19

    # insert last element before eviction
    cache.get_or_set("x", lambda: "x")
    assert len(cache._mru) == 20
    assert len(cache._cache) == 20

    # one more - this should trigger eviction
    cache.get_or_set("y", lambda: "y")
    assert len(cache._mru) == 10
    assert len(cache._cache) == 10

    # now make sure the right entries were evicted
    for x in range(10):
        assert x not in cache._cache
        assert x not in cache._mru

    # validate invariants
    assert cache._cache.keys() == cache._mru.keys()


@pytest.mark.parametrize(
    "expression,result",
    [
        ("[{ key: 1 }]|mapby('key')", [1]),
        ("[{ otherkey: 1 }]|mapby('key')", [None]),
        ("[]|mapby('key')", []),
        ("[{ a: 1, b: 2 }]|mapby('a', 'b', 'c')", [[1, 2, None]]),
        ("0|mapby('key')", None),
    ],
)
def test_mapby_operator(expression, result):
    assert JEXL().evaluate(expression) == result


@pytest.mark.parametrize(
    "expression,result",
    [
        ("[1,2] intersects [2,3]", True),
        ("[1,2] intersects [3,4]", False),
        ("[] intersects []", False),
        ("[1] intersects []", False),
        ("['foo'] intersects ['bar', 'bazz']", False),
        ("['foo'] intersects ['foo', 'foo']", True),
        ("[1] intersects [1] && [2] intersects [2]", True),
        ("[2] intersects [1] + [2]", True),
    ],
)
def test_intersects_operator(expression, result):
    assert JEXL().evaluate(expression) == result


@pytest.mark.parametrize(
    "expression,result",
    [
        ("[1, 2, 3]|min", 1),
        ("['a', 'b']|min", None),
        ("[1, 2, '3']|min", 1),
        ("['a', None]|min", None),
        ("[1, 2, 3]|max", 3),
        ("['a', 'b']|max", None),
        ("[1, 2, '3']|max", 2),
        ("['a', None]|max", None),
        ("[1, 2, 3]|sum", 6),
        ("['a', 'b']|sum", 0),
        ("[1, 2, '3']|sum", 3),
        ("['a', None]|sum", 0),
        ("1.2|ceil", 2),
        ("18.00001|ceil", 19),
        ("None|ceil", None),
        ("'a'|ceil", None),
        ("[1.2]|ceil", None),
        ("1.2|floor", 1),
        ("18.00001|floor", 18),
        ("None|floor", None),
        ("'a'|floor", None),
        ("[1.2]|floor", None),
        ("[1]|avg", 1),
        ("[2, 3]|avg", 2.5),
        ("[2, 3, None, 'foo']|avg", 2.5),
        ("[None, 'foo']|avg", None),
        ("1.49|round", 1),
        ("1.51|round", 2),
        ("1.49|round(1)", 1.5),
        ("1.49|round(2)", 1.49),
        ("1.4445|round(3)", 1.445),
        ("-0.5|round", 0),
        ("-1.14|round(1)", -1.1),
        ("-1.15|round(1)", -1.1),
        ("-1.16|round(1)", -1.2),
        ("'foo'|round", None),
        ("'foo'|round(1)", None),
        ("['foo']|round", None),
        ("['foo']|round(1)", None),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')", [1.4, 2.6]),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|min", 1.4),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|max", 2.6),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|min|round", 1),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|max|round", 3),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|max|round(1)", 2.6),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|sum", 4),
        ("[{ key: 1.4 }, {key: 2.6}]|mapby('key')|sum", 4),
        ("[[{ key: 1.4 }, {key: 2.6}]|mapby('key')|sum]|round", None),
    ],
)
def test_math_transforms(expression, result):
    assert JEXL().evaluate(expression) == result


@pytest.mark.parametrize(
    "expression,result",
    [
        ("[['test1', 'test2'], [1,2]]|stringify", '[["test1", "test2"], [1, 2]]'),
        ("1|stringify", "1"),
    ],
)
def test_stringify_transform(expression, result):
    assert JEXL().evaluate(expression) == result


@pytest.mark.parametrize(
    "expression,expected",
    [
        ("1 + 1", [1, 1]),
        ('["1", 2, 3]', ["1", 2, 3]),
        ('{key: ["foo", "bar"]}', ["foo", "bar"]),
        ("[1|round, [2, 3]|min]|avg", [1, 2, 3]),
    ],
)
def test_caluma_analyzer(expression, expected):
    class NodeAnalyzer(CalumaAnalyzer):
        def visit_Literal(self, node):
            yield node.value

    jexl = JEXL()
    assert list(jexl.analyze(expression, NodeAnalyzer)) == expected



import inspect

from django.core.exceptions import ImproperlyConfigured

from .collections import list_duplicates


def filter_queryset_for(node):
    """Decorate function to define filtering of queryset of specific node."""

    def decorate(fn):
        if not hasattr(fn, "_visibilities"):
            fn._visibilities = []

        fn._visibilities.append(node)
        return fn

    return decorate


class BaseVisibility(object):
    """Basic visibility classes to be extended by any visibility implementation.

    In combination with the decorator `@filter_queryset_for` a custom visibility class
    can define filtering on basis of nodes and its subclasses.

    A custom visibility class could look like this:
    ```
    >>> from caluma.types import Node
    ... from caluma.caluma_form.schema import Form
    ...
    ... class CustomVisibility(BaseVisibility):
    ...     @filter_queryset_for(Node)
    ...     def filter_queryset_for_all(self, node, queryset, info):
    ...         return queryset.filter(created_by_user=info.context.request.user.username)
    ...
    ...     @filter_queryset_for(Form)
    ...     def filter_queryset_for_form(self, node, queryset, info):
    ...         return queryset.exclude(slug='protected-form')
    """

    def __init__(self):
        queryset_fns = inspect.getmembers(self, lambda m: hasattr(m, "_visibilities"))

        queryset_nodes = [
            node.__name__ for _, fn in queryset_fns for node in fn._visibilities
        ]
        queryset_nodes_dups = list_duplicates(queryset_nodes)
        if queryset_nodes_dups:
            raise ImproperlyConfigured(
                f"`filter_queryset_for` defined multiple times for "
                f"{', '.join(queryset_nodes_dups)} in {str(self)}"
            )

        self._filter_querysets_for = {
            node: fn for _, fn in queryset_fns for node in fn._visibilities
        }

    def filter_queryset(self, node, queryset, info):
        for cls in node.mro():
            if cls in self._filter_querysets_for:
                return self._filter_querysets_for[cls](node, queryset, info)

        return queryset


class Any(BaseVisibility):
    """No restrictions, all nodes are exposed."""

    pass


class Union(BaseVisibility):
    """Union result of a list of configured visibility classes."""

    visibility_classes = []

    def filter_queryset(self, node, queryset, info):
        result_queryset = None
        for visibility_class in self.visibility_classes:
            class_result = visibility_class().filter_queryset(node, queryset, info)
            if result_queryset is None:
                result_queryset = class_result
            else:
                result_queryset = result_queryset.union(class_result)

        if result_queryset is not None:
            queryset = queryset.filter(pk__in=result_queryset.values("pk"))

        return queryset



# Generated by Django 2.2.8 on 2020-01-13 07:27

import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0026_auto_20191031_0834")]

    operations = [
        migrations.RemoveIndex(model_name="answer", name="form_answer_date_6485f0_idx"),
        migrations.RemoveIndex(model_name="answer", name="form_answer_meta_98edaa_gin"),
        migrations.RemoveIndex(
            model_name="document", name="form_docume_meta_333f0b_gin"
        ),
        migrations.RemoveIndex(model_name="form", name="form_form_meta_a2eed1_gin"),
        migrations.RemoveIndex(model_name="option", name="form_option_meta_c09f31_gin"),
        migrations.RemoveIndex(
            model_name="question", name="form_questi_meta_ea1ba0_gin"
        ),
        migrations.AddIndex(
            model_name="answer",
            index=models.Index(fields=["date"], name="caluma_form_date_c923c1_idx"),
        ),
        migrations.AddIndex(
            model_name="answer",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta", "value"], name="caluma_form_meta_32fc37_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="document",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_form_meta_89b894_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="form",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_form_meta_796a95_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="option",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_form_meta_e0f506_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="question",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="caluma_form_meta_e7f451_gin"
            ),
        ),
    ]



import json
from tempfile import NamedTemporaryFile

import pytest
from django.core.management import CommandError, call_command

from ..management.commands.dump_caluma_config import get_form_filters

FORM_MODELS = get_form_filters([]).keys()


def sort_dump(dump):
    return sorted(json.loads(dump), key=lambda x: x["pk"])


def test_dump_caluma_config(db, form_and_document, snapshot, form_factory):
    temp_file = NamedTemporaryFile()

    # first: no content
    call_command(
        "dump_caluma_config",
        "test-slug",
        model="caluma_form.Form",
        output=temp_file.name,
    )
    assert not json.loads(temp_file.read())

    form, document, questions_dict, answers_dict = form_and_document(
        use_table=True, use_subform=True
    )

    call_command("dumpdata", *FORM_MODELS, output=temp_file.name)

    temp_file.seek(0)
    dumpdata = sort_dump(temp_file.read())

    # remove non-default answers
    dumpdata = [
        d
        for d in dumpdata
        if d["model"] != "caluma_form.answer" or not d["fields"]["document"]
    ]

    form_factory(slug="unrelated")

    call_command(
        "dump_caluma_config", form.pk, model="caluma_form.Form", output=temp_file.name
    )

    temp_file.seek(0)
    dumpcaluma = sort_dump(temp_file.read())

    assert dumpdata == dumpcaluma
    assert not any(
        (obj["model"], obj["pk"]) == ("caluma_form.form", "unrelated")
        for obj in dumpcaluma
    )


@pytest.mark.parametrize(
    "args,kwargs",
    [
        ([], {"model": "caluma_form.Form"}),
        (["foo"], {"format": "jpeg", "model": "caluma_form.Form"}),
        (["foo"], {}),
    ],
)
def test_dump_caluma_config_error(args, kwargs):
    with pytest.raises(CommandError):
        call_command("dump_caluma_config", *args, **kwargs)



"""Hierarchical representation of a document / form."""
import weakref
from functools import singledispatch
from typing import List, Optional

from .models import AnswerDocument, Question


def object_local_memoise(method):
    def new_method(self, *args, **kwargs):
        if not hasattr(self, "_memoise"):
            self._memoise = {}

        key = str([args, kwargs, method])
        if key in self._memoise:
            return self._memoise[key]
        ret = method(self, *args, **kwargs)
        self._memoise[key] = ret
        return ret

    return new_method


class Element:
    aliases = {}

    def __init__(self, parent=None):
        self._parent = weakref.ref(parent) if parent else None

    def parent(self):
        return self._parent() if self._parent else None

    def children(self):  # pragma: no cover
        return []

    def root(self):
        parent = self.parent()
        if parent:
            return parent.root()
        return self

    def get(self, name, default=None):
        name = self.aliases.get(name, name)

        out = getattr(self, name)

        # if a method is requested, execute it before continuing
        if callable(out):
            out = out()
        if isinstance(out, Element) or isinstance(out, dict):
            return out
        return str(out)


class Field(Element):
    def __init__(self, document, form, question, answer=None, parent=None):
        super().__init__(parent)
        self.document = document
        self.form = form
        self.question = question
        self.answer = answer

    @classmethod
    def factory(cls, document, form, question, answer=None, parent=None):
        if question.type == Question.TYPE_FORM:
            return FieldSet(
                document, form=question.sub_form, question=question, parent=parent
            )
        elif question.type == Question.TYPE_TABLE:
            return RowField(
                document,
                form=question.row_form,
                question=question,
                answer=answer,
                parent=parent,
            )

        return Field(document, form, question, answer, parent=parent)

    def value(self):
        if self.answer is None:
            # no answer object at all - return empty in every case
            return self.question.empty_value()

        elif self.answer.value is not None:
            return self.answer.value

        elif self.question.type == Question.TYPE_TABLE:
            return [
                {cell.question.slug: cell.value() for cell in row.children()}
                for row in self.children()
            ]

        elif self.question.type == Question.TYPE_FILE:
            return self.answer.file

        elif self.question.type == Question.TYPE_DATE:
            return self.answer.date

        elif self.question.type in (
            Question.TYPE_MULTIPLE_CHOICE,
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
        ):
            return []

        # no value, no special handling
        return None

    def __repr__(self):
        return f"<Field question={self.question.slug}, value={self.value()} hidden=({self.question.is_hidden}) req=({self.question.is_required})>"


class RowField(Field):
    @object_local_memoise
    def children(self):
        if not self.answer:
            return []  # pragma: no cover

        rows = AnswerDocument.objects.filter(answer=self.answer).order_by("sort")
        return [
            FieldSet(
                ans_doc.document,
                self.question.row_form,
                question=self.question,
                parent=self.parent(),
            )
            for ans_doc in rows
        ]


class FieldSet(Element):
    aliases = {"formMeta": "form_meta"}

    def __init__(self, document, form, question=None, parent=None):
        super().__init__(parent)
        self.document = document
        self.form = form
        self.form_meta = form.meta
        self.question = question
        self._fields = None
        self._sub_forms = None

    @property
    def fields(self):
        if self._fields is None:
            self._fields = {field.question.slug: field for field in self.children()}
        return self._fields

    @property
    def sub_forms(self) -> List[Field]:
        if self._sub_forms is None:
            self._sub_forms = [
                field
                for field in self.children()
                if field.question.type == Question.TYPE_FORM
            ] + [
                child
                for field in self.children()
                for child in field.children()
                if field.question.type == Question.TYPE_TABLE
            ]
        return self._sub_forms

    def get_field(
        self, question_slug: str, check_parent: bool = True
    ) -> Optional[Field]:
        """Collect fields where the question occurs throughout this structure.

        Cases:
        0. question not in structure
        1. question is in the same form (-> greedily returns the question)
        2. question in a neighbor form, ie. answer would be in same document (excluding tables)
        3. question in multiple neighbor forms
        4. question in a table form (same fieldset)
        5. (question in a table form (different row))
        6. question in a table form, lower than current fieldset
        7. question in upper structure (from table row)

        Expected:
        0: return []
        1-3: answer exists once, but might be in multiple forms -> multiple fields
        4: return only row-local fields (not looking up / other rows)
        5: incomplete row missing answer -> return fields with empty value (like case 4)
        6: return all fields for all rows
        7: same as 1-3
        """

        field = self.fields.get(question_slug)

        if field:
            return field

        elif check_parent:
            field = self.parent().get_field(question_slug) if self.parent() else None
            if field:
                return field

        # OK start looking in subforms / row forms below our level.
        # Since we're looking down, we're disallowing recursing to outer context
        # to avoid recursing back to where we are
        for subform in self.sub_forms:
            sub_field = subform.get_field(question_slug, check_parent=False)
            if sub_field:
                return sub_field

        # if we reach this line, we didn't find the question
        return None

    @object_local_memoise
    def children(self):
        answers = {ans.question_id: ans for ans in self.document.answers.all()}
        return [
            Field.factory(
                document=self.document,
                form=self.form,
                question=question,
                answer=answers.get(question.slug),
                parent=self,
            )
            for question in self.form.questions.all()
        ]

    def __repr__(self):
        q_slug = self.question.slug if self.question else None
        if q_slug:
            return f"<FieldSet fq={q_slug}, doc={self.document.pk} hidden=({self.question.is_hidden}) req=({self.question.is_required})>"

        return f"<FieldSet form={self.form.slug}, doc={self.document.pk}>"


def print_document_structure(document):  # pragma: no cover
    """Print a document's structure.

    Intended halfway as an example on how to use the structure
    classes, and a debugging utility.
    """
    ind = {"i": 0}

    @singledispatch
    def visit(vis):
        raise Exception(f"generic visit(): {vis}")

    @visit.register(Element)
    def _(vis):
        print("   " * ind["i"], vis)
        ind["i"] += 1
        for c in vis.children():
            visit(c)
        ind["i"] -= 1

    struc = FieldSet(document, document.form)
    visit(struc)



import pytest
from django.urls import reverse

from .. import models


@pytest.mark.parametrize(
    "query,variables,operation,operation_name,selection,status_code",
    [
        (
            """
            query foo {
              allCases {
                edges {
                  node {
                    id
            }}}}
            """,
            "",
            "query",
            "foo",
            "allCases",
            200,
        ),
        (
            """
            query {
              allCases {
                edges {
                  node {
                    id
            }}}}
            """,
            "",
            "query",
            None,
            "allCases",
            200,
        ),
        (
            """
            query {
              foo {
                bar
            }}
            """,
            "",
            "query",
            None,
            "foo",
            400,
        ),
        (
            """
            mutation SaveTextQuestion($input: SaveTextQuestionInput!) {
              saveTextQuestion(input: $input) {
                question {
                  id
                  }
                }
            }
            """,
            {"input": {"slug": "some-text-question", "label": "some text question"}},
            "mutation",
            "SaveTextQuestion",
            "saveTextQuestion",
            200,
        ),
        (
            """
            {
            }""",
            "",
            "",
            None,
            None,
            400,
        ),
    ],
)
def test_access_log(
    db,
    snapshot,
    client,
    schema_executor,
    query,
    variables,
    operation,
    operation_name,
    selection,
    status_code,
):
    url = reverse("graphql")
    res = schema_executor(query, variable_values=variables)

    result = client.post(
        url, {"query": query, "variables": variables}, content_type="application/json"
    )

    # schema executor and actual api call should get both success / errors
    assert (
        result.status_code < 400
        and not res.errors
        or result.status_code >= 400
        and res.errors
    )

    assert models.AccessLog.objects.all().count() == 1
    log_entry = models.AccessLog.objects.get()

    assert log_entry.query == query
    assert log_entry.variables == variables
    assert log_entry.operation_name == operation_name
    assert log_entry.operation == operation
    assert log_entry.selection == selection
    assert log_entry.status_code == status_code == result.status_code



from ..relay import extract_global_id


def test_inversible_with_vars(db, schema_executor, document_factory):
    nope, found, excluded = (
        document_factory(meta={}),
        document_factory(meta={"foo": "bar"}),
        document_factory(meta={"foo": "bar", "blah": "blub"}),
    )

    variables = {
        "filter": [
            {"metaValue": [{"key": "foo", "value": "bar"}]},
            {"metaValue": [{"key": "blah", "value": "blub"}], "invert": True},
        ]
    }

    query = """
        query AllDocumentsQuery($filter: [DocumentFilterSetType]) {
          allDocuments(filter: $filter) {
            edges {
              node {
                id
              }
            }
          }
        }
    """

    result = schema_executor(query, variable_values=variables)
    assert not result.errors

    result_ids = [
        extract_global_id(n["node"]["id"]) for n in result.data["allDocuments"]["edges"]
    ]

    assert str(nope.pk) not in result_ids
    assert str(found.pk) in result_ids
    assert str(excluded.pk) not in result_ids


def test_inversible_inline(db, schema_executor, document_factory):
    nope, found, excluded = (
        document_factory(meta={}),
        document_factory(meta={"foo": "bar"}),
        document_factory(meta={"foo": "bar", "blah": "blub"}),
    )
    query = """
        query {
          allDocuments(filter: [
            {metaValue: [{key: "foo", value: "bar"}]},
            {metaValue: [{key: "blah", value: "blub"}], invert: true},
          ]) {
            edges {
              node {
                id
              }
            }
          }
        }
    """

    result = schema_executor(query)
    assert not result.errors

    result_ids = [
        extract_global_id(n["node"]["id"]) for n in result.data["allDocuments"]["edges"]
    ]

    assert str(nope.pk) not in result_ids
    assert str(found.pk) in result_ids
    assert str(excluded.pk) not in result_ids



# Generated by Django 2.2.3 on 2019-07-29 09:08

import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0013_auto_20190718_1235")]

    operations = [
        migrations.AddIndex(
            model_name="case",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="workflow_ca_meta_08775a_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="task",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="workflow_ta_meta_6864a4_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workflow",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="workflow_wo_meta_2ac517_gin"
            ),
        ),
        migrations.AddIndex(
            model_name="workitem",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["meta"], name="workflow_wo_meta_2704a2_gin"
            ),
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-20 09:10

import logging

import django.db.models.deletion
from django.db import migrations, models


def cleanup_invalid_family_pointers(apps, schema_editor):
    Document = apps.get_model("caluma_form.Document")
    invalid_family_pointers = Document.objects.exclude(
        family__in=Document.objects.all().values("pk")
    )
    for doc in invalid_family_pointers:
        logging.warning(
            f"Document pk={doc.pk} (form={doc.form_id}) missing it's family={doc.family}. Resetting to itself"
        )
        doc.family = doc.pk
        doc.save()


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0030_auto_20200219_1359")]

    operations = [
        migrations.RunPython(
            cleanup_invalid_family_pointers, migrations.RunPython.noop
        ),
        migrations.AlterField(
            model_name="document",
            name="family",
            field=models.ForeignKey(
                help_text="Family id which document belongs too.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Document",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldocument",
            name="family",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                help_text="Family id which document belongs too.",
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="caluma_form.Document",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-19 13:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0016_auto_20190510_0843")]

    operations = [
        migrations.AddIndex(
            model_name="answer",
            index=models.Index(fields=["value"], name="form_answer_value_70b2d0_idx"),
        ),
        migrations.AddIndex(
            model_name="answer",
            index=models.Index(fields=["date"], name="form_answer_date_6485f0_idx"),
        ),
    ]



import pytest

from ..collections import list_duplicates


@pytest.mark.parametrize("iterator,expected", [([1, 2, 3], set()), ([1, 1, 2, 3], {1})])
def test_list_duplicates(iterator, expected):
    assert list_duplicates(iterator) == expected



import json

from graphene.utils.str_converters import to_camel_case
from graphql_relay import to_global_id
from rest_framework import fields


def extract_serializer_input_fields(serializer_class, instance):
    serializer = serializer_class(instance)

    result = {}
    for key, value in serializer.data.items():
        field = serializer.fields[key]

        if isinstance(field, fields.JSONField) and key == "meta":
            value = json.dumps(value)
        result[to_camel_case(key)] = value

    result["clientMutationId"] = "testid"

    return result


def extract_global_id_input_fields(instance):
    global_id = to_global_id(type(instance).__name__, instance.pk)
    return {"id": global_id, "clientMutationId": "testid"}



from base64 import b64encode

import pytest

from ..relay import extract_global_id


@pytest.mark.parametrize(
    "id,expected",
    [
        ("slug", "slug"),
        ("V29ya2Zsb3c6bGlzdGVuZWludHJhZw==", "listeneintrag"),
        ("b106", "b106"),
        (b64encode(b"NonExistantModel:foobar"),) * 2,
    ],
)
def test_extract_global_id(id, expected):
    assert extract_global_id(id) == expected



from faker import Faker
from faker.providers.date_time import Provider


class MultilangProvider(Provider):
    """
    Create dictionary compatible with `LocalizedField`.

    A value with given `faker_provider` is created for given languages.
    """

    def multilang(self, faker_provider, languages=("en", "de"), **kwargs):
        value = {}
        for language in languages:
            value[language] = getattr(Faker(language), faker_provider)(*kwargs)

        return value



# Generated by Django 2.2.3 on 2019-07-08 09:05

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "caluma_form",
            "0020_historicalanswer_historicalanswerdocument_historicaldocument_historicalfile_historicalform_historica",
        )
    ]

    operations = [
        migrations.AlterField(
            model_name="form",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this form has been copied from",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
        migrations.AlterField(
            model_name="option",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this option has been copied from",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="caluma_form.Option",
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="row_form",
            field=models.ForeignKey(
                blank=True,
                help_text="Form that represents rows of a TableQuestion",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this question has been copied from",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="caluma_form.Question",
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="sub_form",
            field=models.ForeignKey(
                blank=True,
                help_text="Form referenced in a FormQuestion",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
    ]



from django.core.management.base import BaseCommand
from simple_history.models import registered_models

# Instances which are created at run-time, referencing config-time models
# eg. Answers are created at run-time by users and referencing Question
# NOTE: This checks indirectly if not-nullable referenced objects exist.
# Theoretically it doesn't matter which field to check for, but localized fields seem flaky.
RELATED_HISTORICAL_MODELS = {
    "Answer": {"question__type__isnull": True},
    "Document": {"form__is_published__isnull": True},
    "DynamicOption": {"question__type__isnull": True},
}


class Command(BaseCommand):
    """Cleanup historical records."""

    help = "Cleanup historical records."

    def add_arguments(self, parser):
        parser.add_argument("--force", dest="force", default=False, action="store_true")
        parser.add_argument(
            "-d",
            "--dangling",
            dest="dangling",
            default=False,
            action="store_true",
            help=(
                "Remove records missing a related instance, e.g."
                "HistoricalAnswer missing the Question it relates to."
            ),
        )

    def handle(self, *args, **options):
        force = options["force"]
        dangling = options["dangling"]

        for model in registered_models.values():
            if dangling and model.__name__ in RELATED_HISTORICAL_MODELS:
                qs = model.history.filter(**RELATED_HISTORICAL_MODELS[model.__name__])
            else:
                qs = model.history.none()

            action_str = "Deleting" if force else "Would delete"
            self.stdout.write(
                f'{action_str} {qs.count()} historical records from model "{model.__name__}"'
            )
            if force:
                qs.delete()



import pytest

from caluma.caluma_core.events import SignalHandlingError, on
from caluma.caluma_workflow.api import skip_work_item
from caluma.caluma_workflow.events import (
    post_complete_case,
    post_complete_work_item,
    post_skip_work_item,
)
from caluma.caluma_workflow.models import WorkItem


def test_events(db, work_item_factory, schema_executor):
    work_item = work_item_factory(status=WorkItem.STATUS_READY, child_case=None)

    @on(post_complete_work_item)
    def complete_work_item_event_receiver(sender, work_item, **kwargs):
        work_item.meta = {"been-there": "done that"}
        work_item.save()
        raise Exception("this should not cause an error")

    @on(post_complete_case)
    def complete_case_event_receiver(sender, case, **kwargs):
        case.meta = {"been-there": "done that"}
        case.save()

    query = """
        mutation CompleteWorkItem($input: CompleteWorkItemInput!) {
          completeWorkItem(input: $input) {
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    work_item.refresh_from_db()
    assert work_item.meta == {"been-there": "done that"}
    assert work_item.case.meta == {"been-there": "done that"}


def test_skip_event(db, work_item_factory, schema_executor):
    work_item = work_item_factory(status=WorkItem.STATUS_READY, child_case=None)

    @on(post_complete_work_item)
    def complete_work_item_event_receiver(sender, work_item, **kwargs):
        raise AssertionError(
            "complete event handler should not have been called!"
        )  # pragma: no cover

    @on(post_skip_work_item)
    def skip_work_item_event_receiver(sender, work_item, **kwargs):
        work_item.meta = {"been-there": "done that"}
        work_item.save()

    query = """
        mutation SkipWorkItem($input: SkipWorkItemInput!) {
          skipWorkItem(input: $input) {
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": work_item.pk}}
    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    work_item.refresh_from_db()
    assert work_item.meta == {"been-there": "done that"}


@pytest.mark.parametrize(
    "event,raise_exception",
    [
        (post_skip_work_item, False),
        (post_skip_work_item, True),
        ([post_skip_work_item], True),
    ],
)
def test_event_raise_exceptions(
    db, admin_user, work_item_factory, event, raise_exception
):
    work_item = work_item_factory(status=WorkItem.STATUS_READY, child_case=None)

    @on(event, raise_exception=raise_exception)
    def receiver1(sender, **kwargs):
        raise AssertionError("Error 1")

    @on(event, raise_exception=raise_exception)
    def receiver2(sender, **kwargs):
        raise AssertionError("Error 2")

    if raise_exception:
        with pytest.raises(SignalHandlingError) as exc:
            skip_work_item(work_item, admin_user)

        assert str(exc.value) == "Error 1, Error 2"
    else:
        assert skip_work_item(work_item, admin_user)



import logging
from warnings import warn

logger = logging.getLogger(__name__)


class SignalHandlingError(RuntimeError):
    def __init__(self, exceptions, *args, **kwargs):
        self.caught_exceptions = exceptions

        super().__init__(*args, **kwargs)

    def __str__(self):
        return ", ".join([str(e) for e in self.caught_exceptions])


def on(event, raise_exception=False, *args, **kwargs):
    deprecation_reason = getattr(event, "_deprecation_reason", None)

    if deprecation_reason:  # pragma: no cover
        warn(deprecation_reason, DeprecationWarning)

    def _decorator(func):
        func._raise_exception = raise_exception

        if isinstance(event, (list, tuple)):
            for e in event:
                e.connect(func, **kwargs)
        else:
            event.connect(func, **kwargs)
        return func

    return _decorator


def send_event(signal, **kwargs):
    """
    Send events and handle exceptions in receiver functions.

    This wrapper handles the sending of signals as well as handling of any
    Exceptions that occur. Exceptions are logged with `ERROR` level per
    default, but not raised, in order to prevent rollback of the transaction.
    However, if the `raise_exception` flag on the receiver is set to `True`
    it will collect those errors and raise a `SignalHandlingError` containing
    all caught exceptions.
    """
    result = signal.send_robust(**kwargs)

    caught_exceptions = []
    for func, exc in result:
        if exc is not None:
            if func._raise_exception:
                caught_exceptions.append(exc)

            logger.error(
                f'Error in event handler "{func.__module__}.{func.__name__}":',
                exc_info=exc,
            )

    if caught_exceptions:
        raise SignalHandlingError(caught_exceptions)

    return result


class SendEventSerializerMixin:
    def send_event(self, event, **kwargs):
        send_event(
            event,
            sender=self.context["mutation"],
            user=self.context["request"].user,
            context=getattr(self, "context_data", None),
            **kwargs,
        )  # noqa



# Generated by Django 2.2.3 on 2019-07-29 14:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0022_add_gin_index_to_jsonfields")]

    operations = [
        migrations.AddField(
            model_name="historicaloption",
            name="is_archived",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="option",
            name="is_archived",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-19 13:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0017_auto_20200210_0929")]

    operations = [
        migrations.AlterField(
            model_name="historicaltask",
            name="slug",
            field=models.SlugField(max_length=150),
        ),
        migrations.AlterField(
            model_name="historicalworkflow",
            name="slug",
            field=models.SlugField(max_length=150),
        ),
        migrations.AlterField(
            model_name="task",
            name="slug",
            field=models.SlugField(max_length=150, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="workflow",
            name="slug",
            field=models.SlugField(max_length=150, primary_key=True, serialize=False),
        ),
    ]
